def get_read_data_code(dataset_fullname, collection_name, tag=None, version=None, model=None):
    import_line = 'import dataset_sh as dsh'
    dataset_line = f'with dsh.dataset("{dataset_fullname}")'
    if tag:
        dataset_line += f'.tag("{tag}")'
    elif version:
        dataset_line += f'.version("{version}")'
    else:
        dataset_line += f'.latest()'

    dataset_line += '.open() as reader:'

    print_collections_line = '    # print(reader.collections())'
    if model:
        for_items_line = f'    for item in reader.coll("{collection_name}", model={model})'
    else:
        for_items_line = f'    for item in reader.coll("{collection_name}")'
    for_items_line_print = '        print(item)'
    for_items_line_break = '        break'

    lines = [
        import_line, dataset_line, print_collections_line, for_items_line, for_items_line_print,
        for_items_line_break]
    return '\n'.join(lines)



import json
import math
import os.path
import secrets
import shutil
from datetime import datetime, timedelta
from typing import Optional

from pydantic import BaseModel, Field

from dataset_sh.utils.files import checksum as calculate_checksum


def merge_binary_files(file_list, output_file):
    """
    Merges multiple binary files into a single file.

    :param file_list: List of paths to the binary files to be merged.
    :param output_file: Path to the output file where the merged content will be written.
    """
    with open(output_file, 'wb') as outfile:
        for file_name in file_list:
            with open(file_name, 'rb') as infile:
                outfile.write(infile.read())


def read_file_range(filepath, offset, length):
    with open(filepath, 'rb') as file:
        file.seek(offset)
        file_data = file.read(length)
        return file_data


class FilePartsInfo(BaseModel):
    checksum: str
    parts_count: int
    total_length: int
    chunk_size: int
    secret: str = Field(default_factory=lambda: secrets.token_hex(8))
    expireAt: datetime = Field(default_factory=lambda: datetime.now() + timedelta(hours=1))
    finished: list[int] = Field(default_factory=list)

    # def chunk_size(self, part_id: int) -> int:
    #     if part_id < 0 or part_id >= self.parts_count:
    #         raise ValueError("part_id must be within the range of the total number of parts")
    #
    #     standard_chunk_size = self.total_length // self.parts_count
    #     remainder = self.total_length % self.parts_count
    #
    #     # If it's the last part and there's a remainder, add it to the last chunk's size
    #     if part_id == self.parts_count - 1 and remainder:
    #         return standard_chunk_size + remainder
    #     else:
    #         return standard_chunk_size


class MultipartFileWriter(object):
    """
    A multipart file creator,
    """

    def __init__(self, target, checksum, file_length, chunk_size=1 * 1000 * 1000):
        """

        Args:
            target: the final location of the file
            file_length: length of the file.
        """
        self.target = target
        self.checksum = checksum
        self.file_length = file_length
        self.chunk_size = chunk_size

    @property
    def parts_folder(self):
        return f"{self.target}__parts_folder"

    @property
    def progress_file(self):
        return f"{self.target}._progress"

    def start(self):
        """
        when it starts, it will create a [target]._progress file, and write done the FilePartsInfo object in that file.
        It will also create a [target]._parts folder
        """
        if not os.path.exists(self.parts_folder):
            os.makedirs(self.parts_folder)
        parts_count = int(math.ceil(self.file_length / self.chunk_size))
        info = FilePartsInfo(
            checksum=self.checksum,
            parts_count=parts_count,
            total_length=self.file_length,
            chunk_size=self.chunk_size
        )
        with open(self.progress_file, 'w') as f:
            json.dump(info.model_dump(mode='json'), f)

    def part_files(self) -> list[str]:
        """
        Returns: list of file path of all
        """
        return [os.path.join(self.parts_folder, f"{i}") for i in range(self.load_existing_progress().parts_count)]

    def part_file(self, part_id: int) -> str:
        """
        Returns: path of part part_id, i.e. [target]
        """
        return os.path.join(self.parts_folder, f"{part_id}")

    def complete_marker(self, part_id: int) -> str:
        """
        Returns: path of complete_marker for part part_id
        """
        return os.path.join(self.parts_folder, f"{part_id}.done")

    def load_existing_progress(self) -> Optional[FilePartsInfo]:
        """
        Load the [target]._progress file if exists and return the parsed FilePartsInfo object.
        If the file do not exist yet, return None
        Returns: FilePartsInfo object
        """
        if not os.path.exists(self.progress_file):
            return None
        with open(self.progress_file, 'r') as f:
            data = json.load(f)
        info = FilePartsInfo(**data)
        info.finished = []
        for part_id in range(info.parts_count):
            if os.path.exists(self.complete_marker(part_id)):
                info.finished.append(part_id)

        return info

    def get_completed_parts(self) -> list[int]:
        """
        Returns: list of parts completed, i.e. complete marker exists.
        """
        info = self.load_existing_progress()
        if not info:
            return []
        completed = []
        for part_id in range(info.parts_count):
            if os.path.exists(self.complete_marker(part_id)):
                completed.append(part_id)
        return completed

    def get_uncompleted_parts(self) -> list[int]:
        """
        Returns: list of parts that is missing, i.e. complete marker does not exist.
        """
        info = self.load_existing_progress()
        if not info:
            return []
        uncompleted = []
        for part_id in range(info.parts_count):
            if not os.path.exists(self.complete_marker(part_id)):
                uncompleted.append(part_id)
        return uncompleted

    def mark_file_as_complete(self, part_id: int):
        """
        Returns: create complete marker.
        """
        marker_path = self.complete_marker(part_id)
        with open(marker_path, 'w') as f:
            f.write('')

    def terminate(self):
        '''
        Remove  [target]._progress and all created file parts and markers.
        Returns:

        '''
        if os.path.exists(self.progress_file):
            os.remove(self.progress_file)
        if os.path.exists(self.parts_folder):
            shutil.rmtree(self.parts_folder)

    def done(self):
        """
        Merge all parts file in _parts folder and if the checksum of the final file match the provided value.
        Returns:

        """
        info = self.load_existing_progress()
        if not info:
            raise Exception("No progress info found.")
        merge_binary_files(self.part_files(), self.target)

        if self.checksum == calculate_checksum(self.target):
            self.terminate()
        else:
            raise ValueError('checksum failed.')


class UploadServerResponse(BaseModel):
    action: str
    info: FilePartsInfo


class DownloadServerResponse(BaseModel):
    action: str
    info: FilePartsInfo



import click
from dataset_sh.utils.author import load_author_profile, save_author_profile


@click.group(name="author")
def author_cli():
    """Update default author name and email"""
    pass


@author_cli.command(name="name")
@click.argument("name")
def name_cli(name):
    """Set default author name"""
    cfg = load_author_profile()
    if cfg is None:
        cfg = {}

    author = name
    author_email = cfg.get("author_email")

    save_author_profile(author, author_email)
    click.echo(f"Author name updated to: {author}")


@author_cli.command(name="email")
@click.argument("email")
def email_cli(email):
    """Set default author email"""
    cfg = load_author_profile()
    if cfg is None:
        cfg = {}

    author = cfg.get("author")
    author_email = email

    save_author_profile(author, author_email)
    click.echo(f"Author email updated to: {author_email}")


@author_cli.command(name="show")
def show_cli():
    """Show current author profile"""
    cfg = load_author_profile()
    if cfg:
        author = cfg.get("author", "None")
        author_email = cfg.get("author_email", "None")
        click.echo(f"Author: {author}")
        click.echo(f"Author Email: {author_email}")
    else:
        click.echo("No author profile found.")


import threading
import time

import click

from dataset_sh.server.daemon import daemon_is_running, launch_daemon_app


@click.command(name='gui')
@click.argument('dataset', default='')
@click.option('--host', '-h', 'host', help='if daemon is not running, make the daemon listen to this host.',
              default='localhost')
@click.option('--port', '-p', 'port', help='if daemon is not running, run the daemon on this port.', default=48989)
def gui_cli(dataset, host, port):
    """inspect dataset in web-based gui"""
    host_port = daemon_is_running()
    if host_port:
        host, port = host_port
        url = f'http://{host}:{port}/dataset/{dataset}' if dataset != '' else f'http://{host}:{port}'
        click.launch(url)
    else:
        url = f'http://{host}:{port}/dataset/{dataset}' if dataset != '' else f'http://{host}:{port}'

        def open_browser():
            # Wait a few seconds before opening the browser
            time.sleep(1)
            click.launch(url)

        threading.Thread(target=open_browser).start()
        launch_daemon_app(host, port)



from typing import Optional

from easytype.core import UserDefinedType

from dataset_sh import create
from dataset_sh.utils.author import load_author_profile
from dataset_sh.utils.files import checksum
from dataset_sh.utils.misc import get_tqdm
from dataset_sh.constants import DEFAULT_COLLECTION_NAME


def dump_single_collection(
        fn,
        data,
        name=DEFAULT_COLLECTION_NAME,
        type_annotation: Optional[UserDefinedType] = None,
        description: Optional[str] = None,
        silent=False,
        load_author=True,
):
    """
    Args:
        fn: output file path.
        data: data of the collection.
        name: name of the collection, default to `dataset_sh.constants.DEFAULT_COLLECTION_NAME`
        type_annotation: type annotation for the collection.
        description: describe this version.
        silent: do not report progress if set to True.
    Returns:
    """
    with create(fn) as out:

        if load_author:
            author_info = load_author_profile()
            if author_info:
                author = author_info.get("author", None)
                author_email = author_info.get("author_email", None)

                if author:
                    out.meta.author = author
                if author_email:
                    out.meta.authorEmail = author_email

        if description:
            out.meta.description = description
        out.add_collection(name, data, type_annotation=type_annotation, tqdm=get_tqdm(silent=silent))
    return checksum(fn)


def dump_collections(
        fn,
        data_dict,
        type_dict=None,
        description: Optional[str] = None,
        report_item_progress=False,
        report_collection_progress=False,
        load_author=True,
):
    """
    Args:
        fn: output file path.
        data_dict: data for each collection, dict key is collection name, value is collection data.
        type_dict: type annotation for each collection, dict key is collection name, value is type annotation.
        description: describe this version.
        report_item_progress:
        report_collection_progress:

    Returns:

    """
    if type_dict is None:
        type_dict = {}
    inner_tqdm = get_tqdm(silent=not report_item_progress)
    with create(fn) as out:
        if description:
            out.meta.description = description

        if load_author:
            author_info = load_author_profile()
            if author_info:
                author = author_info.get("author", None)
                author_email = author_info.get("author_email", None)

                if author:
                    out.meta.author = author
                if author_email:
                    out.meta.authorEmail = author_email

        for name, data in data_dict.items():
            if report_collection_progress:
                print(f'Importing collection {name}')
            if len(data) > 0:
                type_annotation = type_dict.get(name, None)
                out.add_collection(name, data, type_annotation=type_annotation, tqdm=inner_tqdm)
    return checksum(fn)



import random


def reservoir_sampling(iterator, k):
    reservoir = []

    for i, item in enumerate(iterator):
        if i < k:
            reservoir.append(item)
        else:
            j = random.randint(0, i)
            if j < k:
                reservoir[j] = item

    return reservoir



import getpass
import os
import json
import warnings
from typing import List, Optional

from pydantic import BaseModel, Field, ValidationError
import dataset_sh.constants as DatasetConstants


class DatasetClientProfile(BaseModel):
    host: str
    key: str
    name: Optional[str] = None


class DatasetClientProfileConfig(BaseModel):
    profiles: List[DatasetClientProfile] = Field(default_factory=list)

    @staticmethod
    def load_profiles():
        config_file = DatasetConstants.CONFIG_JSON
        if os.path.exists(config_file):
            with open(config_file) as fd:
                try:
                    json_config = json.load(fd)
                    ret = DatasetClientProfileConfig(**json_config)
                    return ret
                except (ValidationError, json.decoder.JSONDecodeError):
                    warnings.warn('cannot parse profile config')
        return DatasetClientProfileConfig()

    def get_profile(
            self,
            profile_name='default') -> Optional[DatasetClientProfile]:
        for p in self.profiles:
            if p.name == profile_name:
                return p
        return None

    def get_by_host(self, hostname) -> Optional[DatasetClientProfile]:
        for p in self.profiles:
            if p.host == hostname:
                return p
        return None

    def save(self):
        config_file = DatasetConstants.CONFIG_JSON
        data = self.model_dump(mode='json')
        with open(config_file, 'w') as out:
            json.dump(data, out, indent=4)

    def set_profile(self, host, key, name='default'):
        for p in self.profiles:
            if p.name == name:
                p.host = host
                p.key = key
                return self
        self.profiles.append(DatasetClientProfile(host=host, key=key, name=name))
        return self

    def remove_profile(self, name):
        new_profiles = []
        for p in self.profiles:
            if p.name != name:
                new_profiles.append(p)
        self.profiles = new_profiles
        return self

    @staticmethod
    def read_and_resolve(profile_name):
        return DatasetClientProfileConfig.load_profiles().get_profile(profile_name)

    @staticmethod
    def add_and_save(host, key, name='default'):
        return DatasetClientProfileConfig.load_profiles().set_profile(host, key, name).save()


def ask_override():
    c = input(f'Overwrite existing profile? (y/N)').strip()
    if c.lower() == 'y':
        return True
    return False


def ask_hostname(current):
    hostname = input(f'Enter Host Address: (default: {current})').strip()
    if hostname.strip() == '':
        hostname = current
    return hostname


def ask_name(default_name='default'):
    name = input(f'Profile Name: (default: {default_name})').strip()
    if name == '':
        name = default_name
    return name


def ask_access_key(warn_empty=True):
    key = getpass.getpass('Enter your access key (Your input will be hidden): ').strip()
    return key



import json
import os
from typing import Optional

from dataset_sh.constants import DEFAULT_AUTHOR_JSON


def load_author_profile(author_cfg_path=None) -> Optional[dict]:
    if author_cfg_path is None:
        author_cfg_path = DEFAULT_AUTHOR_JSON
    if os.path.exists(author_cfg_path):
        with open(author_cfg_path) as f:
            return json.load(f)
    else:
        return None


def save_author_profile(author, author_email, author_cfg_path=None):
    if author_cfg_path is None:
        author_cfg_path = DEFAULT_AUTHOR_JSON
    with open(author_cfg_path, 'w') as f:
        json.dump(dict(
            author=author,
            author_email=author_email,
        ), f)



import abc
import os
from typing import Optional


class MarkedFolder:
    location: str
    parent: Optional['MarkedFolder']

    def __init__(self, location, parent=None):
        self.location = location
        self.parent = parent

    @abc.abstractmethod
    def marker_file(self) -> str:
        raise NotImplementedError()

    def post_create(self):
        return

    def exists(self):
        """
        this marked folder exists iff its marker file exist.
        Returns:

        """
        return os.path.isfile(self.marker_file())

    def create_location(self):
        os.makedirs(self.location, exist_ok=True)

    def create(self):
        self.create_location()
        if not self.exists():
            with open(self.marker_file(), 'w') as out:
                out.write('1')
            self.post_create()
        if self.parent and not self.parent.exists():
            self.parent.create()

    def create_if_not_exists(self):
        if not self.exists():
            self.create()
